@@ -17,7 +17,6 @@ from heapq import heapreplace, heapify
 
 import newrelic.packages.six as six
 
-from newrelic.core import config
 from newrelic.core.attribute_filter import DST_ERROR_COLLECTOR
 from newrelic.core.attribute import create_user_attributes
 
@@ -256,6 +255,11 @@ class SampledDataSet(object):
         self.capacity = capacity
         self.num_seen = 0
 
+        if capacity <= 0:
+            def add(*args, **kwargs):
+                self.num_seen += 1
+            self.add = add
+
     @property
     def samples(self):
         return (x[-1] for x in self.pq)
@@ -271,6 +275,9 @@ class SampledDataSet(object):
             'events_seen': self.num_seen
         }
 
+    def __iter__(self):
+        return self.samples
+
     def reset(self):
         self.pq = []
         self.heap = False
@@ -294,7 +301,7 @@ class SampledDataSet(object):
         if priority is None:
             priority = random.random()
 
-        entry = (priority, sample)
+        entry = (priority, self.num_seen, sample)
         if self.num_seen == self.capacity:
             self.pq.append(entry)
             self.heap = self.heap or heapify(self.pq) or True
@@ -307,7 +314,7 @@ class SampledDataSet(object):
             heapreplace(self.pq, entry)
 
     def merge(self, other_data_set):
-        for priority, sample in other_data_set.pq:
+        for priority, seen_at, sample in other_data_set.pq:
             self.add(sample, priority)
 
         # Merge the num_seen from the other_data_set, but take care not to
@@ -316,6 +323,56 @@ class SampledDataSet(object):
         self.num_seen += other_data_set.num_seen - other_data_set.num_samples
 
 
+class LimitedDataSet(list):
+
+    def __init__(self, capacity=200):
+        super(LimitedDataSet, self).__init__()
+
+        self.capacity = capacity
+        self.num_seen = 0
+
+        if capacity <= 0:
+            def add(*args, **kwargs):
+                self.num_seen += 1
+            self.add = add
+
+    @property
+    def samples(self):
+        return self
+
+    @property
+    def num_samples(self):
+        return len(self)
+
+    @property
+    def sampling_info(self):
+        return {
+            'reservoir_size': self.capacity,
+            'events_seen': self.num_seen
+        }
+
+    def should_sample(self):
+        return self.num_seen < self.capacity
+
+    def reset(self):
+        self.clear()
+        self.num_seen = 0
+
+    def add(self, sample):
+        if self.should_sample():
+            self.append(sample)
+        self.num_seen += 1
+
+    def merge(self, other_data_set):
+        for sample in other_data_set:
+            self.add(sample)
+
+        # Merge the num_seen from the other_data_set, but take care not to
+        # double-count the actual samples of other_data_set since the .add
+        # call above will add one to self.num_seen each time
+        self.num_seen += other_data_set.num_seen - other_data_set.num_samples
+
+
 class StatsEngine(object):
 
     """The stats engine object holds the accumulated transactions metrics,
@@ -350,15 +407,14 @@ class StatsEngine(object):
         self.__transaction_events = SampledDataSet()
         self.__error_events = SampledDataSet()
         self.__custom_events = SampledDataSet()
-        self.__span_events = SampledDataSet(self.maximum_allowed_span_events)
+        self.__span_events = SampledDataSet()
         self.__sql_stats_table = {}
         self.__slow_transaction = None
         self.__slow_transaction_map = {}
         self.__slow_transaction_old_duration = None
         self.__slow_transaction_dry_harvests = 0
         self.__transaction_errors = []
-        self.__metric_ids = {}
-        self.__synthetics_events = []
+        self.__synthetics_events = LimitedDataSet()
         self.__synthetics_transactions = []
         self.__xray_transactions = []
         self.xray_sessions = {}
@@ -371,18 +427,6 @@ class StatsEngine(object):
     def stats_table(self):
         return self.__stats_table
 
-    @property
-    def metric_ids(self):
-        """Returns a reference to the dictionary containing the mappings
-        from metric (name, scope) to the integer identifier supplied
-        back from the core application. These integer identifiers are
-        used when sending data to the core application to cut down on
-        the size of data being sent.
-
-        """
-
-        return self.__metric_ids
-
     @property
     def transaction_events(self):
         return self.__transaction_events
@@ -407,37 +451,6 @@ class StatsEngine(object):
     def error_events(self):
         return self.__error_events
 
-    @property
-    def maximum_allowed_span_events(self):
-        if self.__settings is not None:
-            return min(self.__settings.span_events.max_samples_stored,
-                config.SPAN_EVENT_RESERVOIR_SIZE)
-        else:
-            return config.SPAN_EVENT_RESERVOIR_SIZE
-
-    def error_events_sampling_info(self):
-        sampling_info = {
-                'reservoir_size': self.error_events.capacity,
-                'events_seen': self.error_events.num_seen
-        }
-        return sampling_info
-
-    def update_metric_ids(self, metric_ids):
-        """Updates the dictionary containing the mappings from metric
-        (name, scope) to the integer identifier supplied back from the
-        core application. The input should be an iterable returning a
-        list of pairs where first is a dictionary with name and scope
-        keys with corresponding values. The second should be the integer
-        identifier. The dictionary is converted to a (name, scope) tuple
-        for use as key into the internal dictionary containing the
-        mappings.
-
-        """
-
-        for key, value in metric_ids:
-            key = (key['name'], key['scope'])
-            self.__metric_ids[key] = value
-
     def metrics_count(self):
         """Returns a count of the number of unique metrics currently
         recorded for apdex, time and value metrics.
@@ -694,7 +707,7 @@ class StatsEngine(object):
                 'type': 'TransactionError',
                 'error.class': error.type,
                 'error.message': error.message,
-                'timestamp': error.start_time,
+                'timestamp': int(1000.0 * error.start_time),
                 'transactionName': None,
         }
 
@@ -947,11 +960,8 @@ class StatsEngine(object):
         # while transactions from regular requests are saved in another.
 
         if transaction.synthetics_resource_id:
-            if (len(self.__synthetics_events) <
-                    settings.agent_limits.synthetics_events):
-
-                event = transaction.transaction_event(self.__stats_table)
-                self.__synthetics_events.append(event)
+            event = transaction.transaction_event(self.__stats_table)
+            self.__synthetics_events.add(event)
 
         elif (settings.collect_analytics_events and
                 settings.transaction_events.enabled):
@@ -968,8 +978,8 @@ class StatsEngine(object):
         # Merge in span events
 
         if (settings.distributed_tracing.enabled and transaction.sampled and
-                settings.span_events.enabled):
-            for event in transaction.span_events(self.__stats_table):
+                settings.span_events.enabled and settings.collect_span_events):
+            for event in transaction.span_events(self.__settings):
                 self.__span_events.add(event, priority=transaction.priority)
 
     def metric_data(self, normalizer=None):
@@ -1016,10 +1026,7 @@ class StatsEngine(object):
                     list(six.iteritems(normalized_stats)))
 
         for key, value in six.iteritems(normalized_stats):
-            if key not in self.__metric_ids:
-                key = dict(name=key[0], scope=key[1])
-            else:
-                key = self.__metric_ids[key]
+            key = dict(name=key[0], scope=key[1])
             result.append((key, value))
 
         return result
@@ -1232,10 +1239,15 @@ class StatsEngine(object):
             else:
                 force_persist = False
 
+            if trace.include_transaction_trace_request_uri:
+                request_uri = trace.request_uri
+            else:
+                request_uri = None
+
             trace_data.append([transaction_trace.start_time,
                     root.end_time - root.start_time,
                     trace.path,
-                    trace.request_uri,
+                    request_uri,
                     pack_data,
                     trace.guid,
                     None,
@@ -1314,8 +1326,6 @@ class StatsEngine(object):
         self.__slow_transaction_map = {}
         self.__slow_transaction_old_duration = None
         self.__transaction_errors = []
-        self.__metric_ids = {}
-        self.__synthetics_events = []
         self.__synthetics_transactions = []
         self.__xray_transactions = []
         self.xray_sessions = {}
@@ -1324,6 +1334,7 @@ class StatsEngine(object):
         self.reset_error_events()
         self.reset_custom_events()
         self.reset_span_events()
+        self.reset_synthetics_events()
 
     def reset_metric_stats(self):
         """Resets the accumulated statistics back to initial state for
@@ -1360,14 +1371,22 @@ class StatsEngine(object):
             self.__custom_events = SampledDataSet()
 
     def reset_span_events(self):
-        self.__span_events = SampledDataSet(self.maximum_allowed_span_events)
+        if self.__settings is not None:
+            self.__span_events = SampledDataSet(
+                    self.__settings.span_events.max_samples_stored)
+        else:
+            self.__span_events = SampledDataSet()
 
     def reset_synthetics_events(self):
         """Resets the accumulated statistics back to initial state for
         Synthetics events data.
 
         """
-        self.__synthetics_events = []
+        if self.__settings is not None:
+            self.__synthetics_events = LimitedDataSet(
+                    self.__settings.agent_limits.synthetics_events)
+        else:
+            self.__synthetics_events = LimitedDataSet()
 
     def harvest_snapshot(self):
         """Creates a snapshot of the accumulated statistics, error
@@ -1428,13 +1447,13 @@ class StatsEngine(object):
         self.__slow_transaction = None
         self.__transaction_errors = []
         self.__xray_transactions = []
-        self.__synthetics_events = []
         self.__synthetics_transactions = []
 
         self.reset_transaction_events()
         self.reset_error_events()
         self.reset_custom_events()
         self.reset_span_events()
+        self.reset_synthetics_events()
 
         return stats
 
@@ -1445,9 +1464,9 @@ class StatsEngine(object):
 
         """
 
-        stats = StatsEngine()
+        stats = copy.copy(self)
+        stats.reset_stats(self.__settings)
 
-        stats.__settings = self.__settings
         stats.xray_sessions = self.xray_sessions
 
         return stats
@@ -1537,15 +1556,7 @@ class StatsEngine(object):
         # Thus, the events already existing in this object will be newer than
         # those in snapshot, and we favor the newer events.
 
-        if rollback:
-            self.__synthetics_events.extend(snapshot.__synthetics_events)
-        else:
-            if len(snapshot.__synthetics_events) == 1:
-                self.__synthetics_events.append(
-                    snapshot.__synthetics_events[0])
-
-        maximum = self.__settings.agent_limits.synthetics_events
-        self.__synthetics_events = self.__synthetics_events[:maximum]
+        self.__synthetics_events.merge(snapshot.__synthetics_events)
 
     def _merge_error_events(self, snapshot):
 
